# -*- coding: utf-8 -*-
# ==============================================================================
# MIT License
#
# Copyright (c) 2020 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import ABC, abstractmethod
from typing import Optional, Union, Any, Dict, List


class ObjectCoder(ABC):
    """
        Object Coder
        ~~~~~~~~~~~~
        JsON, XML, ...

        1. encode object to string;
        2. decode string to object.
    """

    @abstractmethod
    def encode(self, obj: Any) -> str:
        """
        Encode Map/List object to str

        :param obj: Map or List
        :return: serialized string
        """
        raise NotImplemented

    @abstractmethod
    def decode(self, string: str) -> Optional[Any]:
        """
        Decode str to Map/List object

        :param string: serialized string
        :return: Map or List
        """
        raise NotImplemented


class JSON:
    coder: ObjectCoder = None

    @staticmethod
    def encode(obj: Any) -> str:
        # assert JSON.coder is not None, 'JSON parser not set yet'
        return JSON.coder.encode(obj=obj)

    @staticmethod
    def decode(string: str) -> Optional[Any]:
        # assert JSON.coder is not None, 'JSON parser not set yet'
        return JSON.coder.decode(string=string)


class MapCoder(ObjectCoder, ABC):
    """ coder for json <=> map """

    # Override
    def encode(self, obj: Dict) -> str:
        return JSON.encode(obj=obj)

    # Override
    def decode(self, string: str) -> Optional[Dict]:
        return JSON.decode(string=string)


class ListCoder(ObjectCoder, ABC):
    """ coder for json <=> list """

    # Override
    def encode(self, obj: List) -> str:
        return JSON.encode(obj=obj)

    # Override
    def decode(self, string: str) -> Optional[List]:
        return JSON.decode(string=string)


class JSONMap:
    coder = MapCoder()

    @staticmethod
    def encode(obj: Any) -> str:
        # assert JSONMap.coder is not None, 'JSONMap parser not set yet'
        return JSONMap.coder.encode(obj=obj)

    @staticmethod
    def decode(string: str) -> Optional[Any]:
        # assert JSONMap.coder is not None, 'JSONMap parser not set yet'
        return JSONMap.coder.decode(string=string)


class JSONList:
    coder = ListCoder()

    @staticmethod
    def encode(obj: Any) -> str:
        # assert JSONList.coder is not None, 'JSONList parser not set yet'
        return JSONList.coder.encode(obj=obj)

    @staticmethod
    def decode(string: str) -> Optional[Any]:
        # assert JSONList.coder is not None, 'JSONList parser not set yet'
        return JSONList.coder.decode(string=string)


#
#   Interfaces
#


def json_encode(obj: Union[Dict, List]) -> str:
    return JSON.encode(obj=obj)


def json_decode(string: str) -> Union[Dict, List, None]:
    return JSON.decode(string=string)



# -*- coding: utf-8 -*-
#
#   Ming-Ke-Ming : Decentralized User Identity Authentication
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from .protocol import *

from .address import BroadcastAddress
from .address import ANYWHERE, EVERYWHERE
from .identifier import Identifier
from .identifier import ANYONE, EVERYONE, FOUNDER
from .factory import AccountGeneralFactory, AccountFactoryManager


name = "MingKeMing"

__author__ = 'Albert Moky'

__all__ = [

    #
    #   Protocol
    #
    'EntityType', 'MetaType',
    'Address', 'AddressFactory',
    'ID', 'IDFactory',
    'Meta', 'MetaFactory',
    # 'TAI',
    'Document', 'DocumentFactory',

    #
    #   Core
    #
    'BroadcastAddress',
    'Identifier',
    'AccountGeneralFactory', 'AccountFactoryManager',

    'ANYWHERE', 'EVERYWHERE', 'ANYONE', 'EVERYONE', 'FOUNDER',

]



# -*- coding: utf-8 -*-
# ==============================================================================
# MIT License
#
# Copyright (c) 2023 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from typing import Optional, Any, List, Dict

from ..types import URI, Converter
from ..types import Mapper
from ..crypto import DecryptKey

from .object import JSONMap
from .encode import TransportableData, TransportableDataFactory
from .file import PortableNetworkFile, PortableNetworkFileFactory


class FormatGeneralFactory:

    def __init__(self):
        super().__init__()
        self.__pnf_factory: Optional[PortableNetworkFileFactory] = None
        # str(algorithm) => TransportableData.Factory
        self.__ted_factories: Dict[str, TransportableDataFactory] = {}

    # noinspection PyMethodMayBeStatic
    def split(self, text: str) -> List[str]:
        """
        Split text string to array: ["{TEXT}", "{algorithm}"]

        :param text: '{TEXT}', or
                     'base64,{BASE64_ENCODE}', or
                     'data:image/png;base64,{BASE64_ENCODE}'
        :return: text + algorithm
        """
        pos1 = text.find('://')
        if pos1 > 0:
            # [URL]
            return [text]
        pos1 = text.find(';') + 1
        pos2 = text.find(',', pos1)
        if pos2 > pos1:
            # [data, algorithm]
            alg = text[pos1:pos2]
            pos2 += 1  # skip ','
            data = text[pos2:]
            return [data, alg]
        # [data]
        if pos1 > 0:
            text = text[pos1:]
        return [text]

    def decode(self, data: Any, default_key: str) -> Optional[Dict]:
        if isinstance(data, Mapper):
            return data.dictionary
        elif isinstance(data, Dict):
            return data
        text = data if isinstance(data, str) else str(data)
        if len(text) == 0:
            return None
        elif text.startswith('{') and text.endswith('}'):
            return JSONMap.decode(string=text)
        array = self.split(text=text)
        if len(array) == 1:
            return {
                default_key: array[0],
            }
        assert len(array) == 2, 'split error: %s => %s' % (text, array)
        return {
            'algorithm': array[1],
            'data': array[0],
        }

    #
    #   TED - Transportable Encoded Data
    #

    # noinspection PyMethodMayBeStatic
    def get_data_algorithm(self, ted: Dict[str, Any], default: Optional[str]) -> Optional[str]:
        value = ted.get('algorithm')
        return Converter.get_str(value=value, default=default)

    def set_transportable_data_factory(self, algorithm: str, factory: TransportableDataFactory):
        self.__ted_factories[algorithm] = factory

    def get_transportable_data_factory(self, algorithm: str) -> Optional[TransportableDataFactory]:
        return self.__ted_factories.get(algorithm)

    def create_transportable_data(self, algorithm: str, data: bytes) -> TransportableData:
        factory = self.get_transportable_data_factory(algorithm=algorithm)
        assert factory is not None, 'data algorithm not support: %s' % algorithm
        return factory.create_transportable_data(data=data)

    def parse_transportable_data(self, ted: Any) -> Optional[TransportableData]:
        if ted is None:
            return None
        elif isinstance(ted, TransportableData):
            return ted
        # unwrap
        info = self.decode(data=ted, default_key='data')
        if info is None:
            # assert False, 'TED error: %s' % ted
            return None
        algorithm = self.get_data_algorithm(info, default='*')
        factory = self.get_transportable_data_factory(algorithm=algorithm)
        if factory is None:
            assert algorithm != '*', 'TED factory not ready'
            factory = self.get_transportable_data_factory(algorithm='*')  # unknown
            assert factory is not None, 'default TED factory not found'
        return factory.parse_transportable_data(info)

    #
    #   PNF - Portable Network File
    #

    def set_portable_network_file_factory(self, factory: PortableNetworkFileFactory):
        self.__pnf_factory = factory

    def get_portable_network_file_factory(self) -> Optional[PortableNetworkFileFactory]:
        return self.__pnf_factory

    def create_portable_network_file(self, data: Optional[TransportableData], filename: Optional[str],
                                     url: Optional[URI], password: Optional[DecryptKey]) -> PortableNetworkFile:
        factory = self.get_portable_network_file_factory()
        assert factory is not None, 'PNF factory not ready'
        return factory.create_portable_network_file(data=data, filename=filename, url=url, password=password)

    def parse_portable_network_file(self, pnf: Any) -> Optional[PortableNetworkFile]:
        if pnf is None:
            return None
        elif isinstance(pnf, PortableNetworkFile):
            return pnf
        # unwrap
        info = self.decode(data=pnf, default_key='URL')
        if info is None:
            # assert False, 'PNF error: %s' % pnf
            return None
        factory = self.get_portable_network_file_factory()
        assert factory is not None, 'PNF factory not ready'
        return factory.parse_portable_network_file(info)


# Singleton
class FormatFactoryManager:

    general_factory = FormatGeneralFactory()



# -*- coding: utf-8 -*-
#
#   Ming-Ke-Ming : Decentralized User Identity Authentication
#
#                                Written in 2021 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2021 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from typing import Optional, Any, List, Dict

from .types import Converter
from .types import Wrapper
from .format import TransportableData
from .crypto import SignKey, VerifyKey

from .protocol import Address, AddressFactory
from .protocol import ID, IDFactory
from .protocol import Meta, MetaFactory
from .protocol import Document, DocumentFactory


class AccountGeneralFactory:

    def __init__(self):
        super().__init__()
        # AddressFactory
        self.__address_factory: Optional[AddressFactory] = None
        # IDFactory
        self.__id_factory: Optional[IDFactory] = None
        # int(type) -> MetaFactory
        self.__meta_factories: Dict[int, MetaFactory] = {}
        # str(type) -> DocumentFactory
        self.__document_factories: Dict[str, DocumentFactory] = {}

    #
    #   Address
    #

    def set_address_factory(self, factory: AddressFactory):
        self.__address_factory = factory

    def get_address_factory(self) -> Optional[AddressFactory]:
        return self.__address_factory

    def generate_address(self, meta: Meta, network: int = None) -> Address:
        factory = self.get_address_factory()
        assert factory is not None, 'address factory not set'
        return factory.generate_address(meta=meta, network=network)

    def create_address(self, address: str) -> Optional[Address]:
        factory = self.get_address_factory()
        assert factory is not None, 'address factory not set'
        return factory.create_address(address=address)

    def parse_address(self, address: Any) -> Optional[Address]:
        if address is None:
            return None
        elif isinstance(address, Address):
            return address
        string = Wrapper.get_str(address)
        if string is None:
            # assert False, 'address error: %s' % address
            return None
        factory = self.get_address_factory()
        assert factory is not None, 'address factory not set'
        return factory.parse_address(address=string)

    #
    #   ID
    #

    def set_identifier_factory(self, factory: IDFactory):
        self.__id_factory = factory

    def get_identifier_factory(self) -> Optional[IDFactory]:
        return self.__id_factory

    def generate_identifier(self, meta: Meta, network: Optional[int], terminal: Optional[str]) -> ID:
        factory = self.get_identifier_factory()
        assert factory is not None, 'ID factory not set'
        return factory.generate_identifier(meta=meta, network=network, terminal=terminal)

    def create_identifier(self, name: Optional[str], address: Address, terminal: Optional[str]) -> ID:
        factory = self.get_identifier_factory()
        assert factory is not None, 'ID factory not set'
        return factory.create_identifier(name=name, address=address, terminal=terminal)

    def parse_identifier(self, identifier: Any) -> Optional[ID]:
        if identifier is None:
            return None
        elif isinstance(identifier, ID):
            return identifier
        string = Wrapper.get_str(identifier)
        if string is None:
            # assert False, 'ID error: %s' % identifier
            return None
        factory = self.get_identifier_factory()
        assert factory is not None, 'ID factory not set'
        return factory.parse_identifier(identifier=string)

    def convert_identifiers(self, array: List[str]) -> List[ID]:
        """
        Convert ID list from string array

        :param array: ID string array
        :return: ID list
        """
        result = []
        for item in array:
            identifier = self.parse_identifier(identifier=item)
            if identifier is None:
                # id error
                continue
            result.append(identifier)
        return result

    # noinspection PyMethodMayBeStatic
    def revert_identifiers(self, array: List[ID]) -> List[str]:
        """
        Revert ID list to string array

        :param array: ID list
        :return: string array
        """
        result = []
        for item in array:
            result.append(str(item))
        return result

    #
    #   Meta
    #

    def set_meta_factory(self, version: int, factory: MetaFactory):
        self.__meta_factories[version] = factory

    def get_meta_factory(self, version: int) -> Optional[MetaFactory]:
        return self.__meta_factories.get(version)

    # noinspection PyMethodMayBeStatic
    def get_meta_type(self, meta: Dict[str, Any], default: Optional[int]) -> Optional[int]:
        """ get meta type(version) """
        value = meta.get('type')
        return Converter.get_int(value=value, default=default)

    def generate_meta(self, version: int, private_key: SignKey,
                      seed: Optional[str]) -> Meta:
        factory = self.get_meta_factory(version)
        assert factory is not None, 'failed to get meta factory: %d' % version
        return factory.generate_meta(private_key, seed=seed)

    def create_meta(self, version: int, public_key: VerifyKey,
                    seed: Optional[str], fingerprint: Optional[TransportableData]) -> Meta:
        factory = self.get_meta_factory(version)
        assert factory is not None, 'failed to get meta factory: %d' % version
        return factory.create_meta(public_key, seed=seed, fingerprint=fingerprint)

    def parse_meta(self, meta: Any) -> Optional[Meta]:
        if meta is None:
            return None
        elif isinstance(meta, Meta):
            return meta
        info = Wrapper.get_dict(meta)
        if info is None:
            # assert False, 'meta error: %s' % meta
            return None
        version = self.get_meta_type(meta=info, default=0)
        assert version > 0, 'meta error: %s' % meta
        factory = self.get_meta_factory(version)
        if factory is None:
            factory = self.get_meta_factory(0)  # unknown
            assert factory is not None, 'default meta factory not found'
        return factory.parse_meta(meta=info)

    #
    #   Document
    #

    def set_document_factory(self, doc_type: str, factory: DocumentFactory):
        self.__document_factories[doc_type] = factory

    def get_document_factory(self, doc_type: str) -> Optional[DocumentFactory]:
        return self.__document_factories.get(doc_type)

    # noinspection PyMethodMayBeStatic
    def get_document_type(self, document: Dict[str, Any], default: Optional[str]) -> Optional[str]:
        value = document.get('type')
        return Converter.get_str(value=value, default=default)

    def create_document(self, doc_type: str, identifier: ID,
                        data: Optional[str], signature: Optional[TransportableData]) -> Document:
        factory = self.get_document_factory(doc_type)
        assert factory is not None, 'document factory not found for type: %s' % doc_type
        return factory.create_document(identifier=identifier, data=data, signature=signature)

    def parse_document(self, document: Any) -> Optional[Document]:
        if document is None:
            return None
        elif isinstance(document, Document):
            return document
        info = Wrapper.get_dict(document)
        if info is None:
            # assert False, 'document error: %s' % document
            return None
        doc_type = self.get_document_type(document=info, default='*')
        factory = self.get_document_factory(doc_type)
        if factory is None:
            assert doc_type != '*', 'document factory not ready: %s' % document
            factory = self.get_document_factory('*')  # unknown
            assert factory is not None, 'default document factory not found'
        return factory.parse_document(document=info)


# Singleton
class AccountFactoryManager:

    general_factory = AccountGeneralFactory()



# -*- coding: utf-8 -*-
# ==============================================================================
# MIT License
#
# Copyright (c) 2023 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import ABC, abstractmethod
from typing import Optional, Any, Dict

from ..types import Mapper


class TransportableData(Mapper, ABC):
    """
        Transportable Data
        ~~~~~~~~~~~~~~~~~~
        TED - Transportable Encoded Data

            0. "{BASE64_ENCODE}"
            1. "base64,{BASE64_ENCODE}"
            2. "data:image/png;base64,{BASE64_ENCODE}"
            3. {
                algorithm : "base64",
                data      : "...",      // base64_encode(data)
                ...
            }
    """

    #
    #   encode algorithm
    #
    DEFAULT = 'base64'
    BASE_64 = 'base64'
    BASE_58 = 'base58'
    HEX = 'hex'

    @property
    @abstractmethod
    def algorithm(self) -> Optional[str]:
        """
        Get data encode algorithm

        :return: 'base64'
        """
        raise NotImplemented

    @property
    @abstractmethod
    def data(self) -> Optional[bytes]:
        """
        Get original data

        :return: plaintext
        """
        raise NotImplemented

    @abstractmethod
    def __str__(self) -> str:
        """
        Get encoded string

        :return: '{BASE64_ENCODE}', or
                 'base64,{BASE64_ENCODE}', or
                 'data:image/png;base64,{BASE64_ENCODE}', or
                 '{...}'
        """
        raise NotImplemented

    @property
    @abstractmethod
    def object(self) -> Any:
        """
        to_json()

        :return: str, or dict
        """
        return NotImplemented

    #
    #  Conveniences
    #

    @classmethod
    def encode(cls, data: bytes) -> Any:
        ted = cls.create(data=data)
        # assert isinstance(ted, TransportableData), 'should not happen'
        return ted.object

    @classmethod
    def decode(cls, ted: Any) -> Optional[bytes]:
        ted = cls.parse(ted)
        if ted is not None:  # isinstance(ted, TransportableData):
            return ted.data

    #
    #  Factory methods
    #

    @classmethod
    def create(cls, data: bytes, algorithm: str = None):  # -> TransportableData;
        if algorithm is None:
            algorithm = cls.DEFAULT
        gf = general_factory()
        return gf.create_transportable_data(data=data, algorithm=algorithm)

    @classmethod
    def parse(cls, ted: Any):  # -> Optional[TransportableData]:
        gf = general_factory()
        return gf.parse_transportable_data(ted)

    @classmethod
    def factory(cls, algorithm: str):  # -> Optional[TransportableDataFactory]:
        gf = general_factory()
        return gf.get_transportable_data_factory(algorithm=algorithm)

    @classmethod
    def register(cls, algorithm: str, factory):
        gf = general_factory()
        gf.set_transportable_data_factory(algorithm=algorithm, factory=factory)


def general_factory():
    from .factory import FormatFactoryManager
    return FormatFactoryManager.general_factory


class TransportableDataFactory(ABC):
    """ TED factory """

    @abstractmethod
    def create_transportable_data(self, data: bytes) -> TransportableData:
        """
        Create TED

        :param data: original data
        :return: TED object
        """
        raise NotImplemented

    @abstractmethod
    def parse_transportable_data(self, ted: Dict[str, Any]) -> Optional[TransportableData]:
        """
        Parse map object to TED

        :param ted: TED info
        :return: TED object
        """
        raise NotImplemented



# -*- coding: utf-8 -*-
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import ABC, abstractmethod

from .cryptography import CryptographyKey


# noinspection PyAbstractClass
class AsymmetricKey(CryptographyKey, ABC):

    RSA = 'RSA'  # -- "RSA/ECB/PKCS1Padding", "SHA256withRSA"
    ECC = 'ECC'


class SignKey(AsymmetricKey, ABC):

    @abstractmethod
    def sign(self, data: bytes) -> bytes:
        """
        signature = sign(data, SK);

        :param data: message data
        :return: signature
        """
        raise NotImplemented


class VerifyKey(AsymmetricKey, ABC):

    @abstractmethod
    def verify(self, data: bytes, signature: bytes) -> bool:
        """
        OK = verify(data, signature, PK)

        :param data:      message data
        :param signature: signature of message data
        :return: True on signature matched
        """
        raise NotImplemented

    @abstractmethod
    def match_sign_key(self, key: SignKey) -> bool:
        """
        OK = verify(data, sign(data, SK), PK)

        :param key: private key
        :return: True on signature matched
        """
        raise NotImplemented



# -*- coding: utf-8 -*-
# ==============================================================================
# MIT License
#
# Copyright (c) 2023 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import ABC, abstractmethod
from typing import Optional, Any, Dict

from ..types import Mapper
from ..types import URI
from ..crypto import DecryptKey

from .encode import TransportableData


class PortableNetworkFile(Mapper, ABC):
    """
        Transportable File
        ~~~~~~~~~~~~~~~~~~
        PNF - Portable Network File

            0. "{URL}"
            1. "base64,{BASE64_ENCODE}"
            2. "data:image/png;base64,{BASE64_ENCODE}"
            3. {
                data     : "...",        // base64_encode(fileContent)
                filename : "avatar.png",

                URL      : "http://...", // download from CDN
                // before fileContent uploaded to a public CDN,
                // it can be encrypted by a symmetric key
                key      : {             // symmetric key to decrypt file content
                    algorithm : "AES",   // "DES", ...
                    data      : "{BASE64_ENCODE}",
                    ...
                }
            }
    """

    #
    #   When file data is too big, don't set it in this dictionary,
    #   but upload it to a CDN and set the download URL instead.
    #
    @property
    @abstractmethod
    def data(self) -> Optional[bytes]:
        raise NotImplemented

    @data.setter
    @abstractmethod
    def data(self, content: Optional[bytes]):
        raise NotImplemented

    @property
    @abstractmethod
    def filename(self) -> Optional[str]:
        raise NotImplemented

    @filename.setter
    @abstractmethod
    def filename(self, string: Optional[str]):
        raise NotImplemented

    #
    #   Download URL
    #
    @property
    @abstractmethod
    def url(self) -> Optional[URI]:
        # download URL from CDN
        raise NotImplemented

    @url.setter
    @abstractmethod
    def url(self, string: Optional[URI]):
        raise NotImplemented

    #
    #   Password for decrypting the downloaded data from CDN,
    #   default is a plain key, which just return the same data when decrypting.
    #
    @property
    @abstractmethod
    def password(self) -> Optional[DecryptKey]:
        raise NotImplemented

    @password.setter
    @abstractmethod
    def password(self, key: Optional[DecryptKey]):
        raise NotImplemented

    @abstractmethod
    def __str__(self) -> str:
        """
        Get encoded string

        :return: 'URL', or
                 'base64,{BASE64_ENCODE}', or
                 'data:image/png;base64,{BASE64_ENCODE}', or
                 '{...}'
        """
        raise NotImplemented

    @property
    @abstractmethod
    def object(self) -> Any:
        """
        to_json()

        :return: str, or dict
        """
        return NotImplemented

    #
    #  Factory methods
    #

    @classmethod
    def create_from_url(cls, url: URI, password: Optional[DecryptKey]):
        return cls.create(url=url, password=password)

    @classmethod
    def create_from_data(cls, data: bytes, filename: Optional[str]):
        ted = TransportableData.create(data=data)
        return cls.create(data=ted, filename=filename)

    @classmethod
    def create(cls, data: Optional[TransportableData] = None, filename: Optional[str] = None,
               url: Optional[URI] = None, password: Optional[DecryptKey] = None):  # -> PortableNetworkFile:
        gf = general_factory()
        return gf.create_portable_network_file(data=data, filename=filename, url=url, password=password)

    @classmethod
    def parse(cls, pnf: Any):  # -> Optional[PortableNetworkFile]:
        gf = general_factory()
        return gf.parse_portable_network_file(pnf)

    @classmethod
    def factory(cls):  # -> Optional[PortableNetworkFileFactory]:
        gf = general_factory()
        return gf.get_portable_network_file_factory()

    @classmethod
    def register(cls, factory):
        gf = general_factory()
        gf.set_portable_network_file_factory(factory=factory)


def general_factory():
    from .factory import FormatFactoryManager
    return FormatFactoryManager.general_factory


class PortableNetworkFileFactory(ABC):
    """ PNF factory """

    @abstractmethod
    def create_portable_network_file(self, data: Optional[TransportableData], filename: Optional[str],
                                     url: Optional[URI], password: Optional[DecryptKey]) -> PortableNetworkFile:
        """
        Create PNF

        :param data:     file data (not encrypted)
        :param filename: file name
        :param url:      download URL
        :param password: decrypt key for downloaded data
        :return: PNF object
        """
        raise NotImplemented

    @abstractmethod
    def parse_portable_network_file(self, pnf: Dict[str, Any]) -> Optional[PortableNetworkFile]:
        """
        Parse map object to PNF

        :param pnf: PNF info
        :return: PNF object
        """
        raise NotImplemented



# -*- coding: utf-8 -*-
#
#   Ming-Ke-Ming : Decentralized User Identity Authentication
#
#                                Written in 2019 by Moky <albert.moky@gmail.com>
#
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import ABC, abstractmethod
from typing import Optional, Any, Dict

from ..types import DateTime
from ..types import Mapper
from ..format import TransportableData

from .identifier import ID
from .tai import TAI


class Document(TAI, Mapper, ABC):
    """
        User/Group Profile
        ~~~~~~~~~~~~~~~~~~
        This class is used to generate entity profile

            data format: {
                ID        : "EntityID",        // entity ID
                type      : "visa",            // "bulletin", ...
                data      : "{JSON}",          // data = json_encode(info)
                signature : "{BASE64_ENCODE}"  // signature = sign(data, SK);
            }
    """

    #
    #  Document types
    #
    VISA = 'visa'          # for user info (communicate key)
    PROFILE = 'profile'    # for user profile (reserved)
    BULLETIN = 'bulletin'  # for group info (owner, assistants)

    @property
    @abstractmethod
    def type(self) -> Optional[str]:
        """
        Get document type

        :return: doc type
        """
        raise NotImplemented

    @property
    @abstractmethod
    def identifier(self) -> ID:
        """
        Get entity ID

        :return: Entity ID
        """
        raise NotImplemented

    @property
    @abstractmethod
    def time(self) -> Optional[DateTime]:
        """
        Get sign time

        :return: timestamp
        """
        raise NotImplemented

    #
    #  properties getter/setter
    #
    @property
    @abstractmethod
    def name(self) -> Optional[str]:
        """
        Get entity name

        :return: name string
        """
        raise NotImplemented

    @name.setter
    @abstractmethod
    def name(self, string: str):
        """
        Set entity name

        :param string: name string
        :return:
        """
        raise NotImplemented

    #
    #   Factory Methods
    #

    @classmethod
    def create(cls, doc_type: str, identifier: ID,
               data: Optional[str] = None, signature: Optional[TransportableData] = None):  # -> Optional[Document]:
        gf = general_factory()
        return gf.create_document(doc_type, identifier=identifier, data=data, signature=signature)

    @classmethod
    def parse(cls, document: Any):  # -> Optional[Document]:
        gf = general_factory()
        return gf.parse_document(document=document)

    @classmethod
    def register(cls, doc_type: str, factory):
        gf = general_factory()
        gf.set_document_factory(doc_type, factory=factory)

    @classmethod
    def factory(cls, doc_type: str):  # -> Optional[DocumentFactory]:
        gf = general_factory()
        return gf.get_document_factory(doc_type)


def general_factory():
    from ..factory import AccountFactoryManager
    return AccountFactoryManager.general_factory


class DocumentFactory(ABC):

    @abstractmethod
    def create_document(self, identifier: ID, data: Optional[str], signature: Optional[TransportableData]) -> Document:
        """
        1. Create a new empty document with entity ID

        2. Create document with data & signature loaded from local storage

        :param identifier: entity ID
        :param data:       document data
        :param signature:  document signature
        :return: Document
        """
        raise NotImplemented

    @abstractmethod
    def parse_document(self, document: Dict[str, Any]) -> Optional[Document]:
        """
        Parse map object to entity document

        :param document:
        :return:
        """
        raise NotImplemented



# -*- coding: utf-8 -*-
# ==============================================================================
# MIT License
#
# Copyright (c) 2024 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import copy
from abc import ABC
from typing import Any, Dict, List

from .wrapper import Mapper


class Copier(ABC):

    @classmethod
    def copy(cls, o: Any) -> Any:
        if o is None:
            return None
        elif isinstance(o, Mapper):
            return cls.copy_map(o.dictionary)
        elif isinstance(o, Dict):
            return cls.copy_map(o)
        elif isinstance(o, List):
            return cls.copy_list(o)
        else:
            return o

    @classmethod
    def copy_map(cls, d: Dict) -> Dict:
        return d.copy()

    @classmethod
    def copy_list(cls, a: List) -> List:
        return a.copy()

    #
    #   Deep Copy
    #

    @classmethod
    def deep_copy(cls, o: Any) -> Any:
        if o is None:
            return None
        elif isinstance(o, Mapper):
            return cls.deep_copy_map(o.dictionary)
        elif isinstance(o, Dict):
            return cls.deep_copy_map(o)
        elif isinstance(o, List):
            return cls.deep_copy_list(o)
        else:
            # return o
            return copy.deepcopy(o)

    @classmethod
    def deep_copy_map(cls, d: Dict) -> Dict:
        dictionary = {}
        for k in d:
            v = d[k]
            dictionary[k] = cls.deep_copy(v)
        return dictionary

    @classmethod
    def deep_copy_list(cls, a: List) -> List:
        array = []
        for item in a:
            array.append(cls.deep_copy(item))
        return array



# -*- coding: utf-8 -*-
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Data Format
    ~~~~~~~~~~~

    UTF-8, JsON, Hex, Base58, Base64, ...
    TED, PNF
"""

from .data import DataCoder, Hex, Base58, Base64
from .data import hex_encode, hex_decode
from .data import base58_encode, base58_decode, base64_encode, base64_decode

from .object import ObjectCoder, JSON
from .object import MapCoder, ListCoder, JSONMap, JSONList
from .object import json_encode, json_decode

from .string import StringCoder, UTF8
from .string import utf8_encode, utf8_decode


from .encode import TransportableData, TransportableDataFactory
from .file import PortableNetworkFile, PortableNetworkFileFactory
from .factory import FormatGeneralFactory, FormatFactoryManager


name = "Crypto"

__author__ = 'Albert Moky'

__all__ = [

    # Data Format
    'DataCoder', 'Hex', 'Base58', 'Base64',
    'ObjectCoder', 'JSON',
    'MapCoder', 'JSONMap',
    'ListCoder', 'JSONList',
    'StringCoder', 'UTF8',

    'hex_encode', 'hex_decode',
    'base58_encode', 'base58_decode',
    'base64_encode', 'base64_decode',
    'json_encode', 'json_decode',
    'utf8_encode', 'utf8_decode',

    # TED, PNF
    'TransportableData', 'TransportableDataFactory',
    'PortableNetworkFile', 'PortableNetworkFileFactory',
    'FormatGeneralFactory', 'FormatFactoryManager',

]



# -*- coding: utf-8 -*-
# ==============================================================================
# MIT License
#
# Copyright (c) 2023 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

import time
from typing import Union, Any

URI = str


class DateTime:

    def __init__(self, timestamp: float):
        super().__init__()
        self.__timestamp = timestamp  # seconds since the Epoch.
        self.__localtime = None

    @property
    def timestamp(self) -> float:
        """ seconds from 1970-01-01 00:00:00 """
        return self.__timestamp

    def before(self, other) -> bool:
        """ Return self<value. """
        if isinstance(other, DateTime):
            other = other.timestamp
        # assert isinstance(other, float), 'other time error: %s' % other
        return self.__timestamp < other

    def after(self, other) -> bool:
        """ Return self>value. """
        if isinstance(other, DateTime):
            other = other.timestamp
        # assert isinstance(other, float), 'other time error: %s' % other
        return self.__timestamp > other

    # Override
    def __lt__(self, other) -> bool:
        """ Return self<value. """
        if not isinstance(other, float):
            other = float(other)
        return self.__timestamp < other

    # Override
    def __gt__(self, other) -> bool:
        """ Return self>value. """
        if not isinstance(other, float):
            other = float(other)
        return self.__timestamp > other

    # Override
    def __le__(self, other) -> bool:
        """ Return self<=value. """
        if not isinstance(other, float):
            other = float(other)
        return self.__timestamp <= other

    # Override
    def __ge__(self, other) -> bool:
        """ Return self>=value. """
        if not isinstance(other, float):
            other = float(other)
        return self.__timestamp >= other

    # Override
    def __eq__(self, other) -> bool:
        """ Return self==value. """
        if not isinstance(other, float):
            other = float(other)
        return self.__timestamp == other

    # Override
    def __ne__(self, other) -> bool:
        """ Return self!=value. """
        if not isinstance(other, float):
            other = float(other)
        return self.__timestamp != other

    # Override
    def __hash__(self) -> int:
        """ Return hash(self). """
        return hash(self.__timestamp)

    # Override
    def __int__(self) -> int:
        """ Return int(self). """
        return int(self.__timestamp)

    # Override
    def __float__(self) -> float:
        """ Return float(self). """
        return self.__timestamp

    # Override
    def __add__(self, other) -> float:
        """ Return self+value. """
        if not isinstance(other, float):
            other = float(other)
        return self.__timestamp + other

    # Override
    def __radd__(self, other) -> float:
        """ Return value+self. """
        if not isinstance(other, float):
            other = float(other)
        return other + self.__timestamp

    # Override
    def __sub__(self, other) -> float:
        """ Return self-value. """
        if not isinstance(other, float):
            other = float(other)
        return self.__timestamp - other

    # Override
    def __rsub__(self, other) -> float:
        """ Return value-self. """
        if not isinstance(other, float):
            other = float(other)
        return other - self.__timestamp

    # Override
    def __str__(self):
        """ Return str(self). """
        return time.strftime('%Y-%m-%d %H:%M:%S', self.localtime)

    # Override
    def __repr__(self) -> str:
        """ Return repr(self). """
        return time.strftime('%a %b %d %H:%M:%S %Y', self.localtime)

    @property
    def localtime(self) -> time.struct_time:
        if self.__localtime is None:
            self.__localtime = time.localtime(self.__timestamp)
        return self.__localtime

    @property
    def year(self) -> int:
        """ year, for example, 1993 """
        return self.localtime.tm_year

    @property
    def month(self) -> int:
        """ month of year, range [1, 12] """
        return self.localtime.tm_mon

    @property
    def day(self) -> int:
        """ day of month, range [1, 31] """
        return self.localtime.tm_mday

    @property
    def weak_day(self) -> int:
        """ day of week, range [0, 6], Monday is 0 """
        return self.localtime.tm_wday

    @property
    def year_day(self) -> int:
        """ day of year, range [1, 366] """
        return self.localtime.tm_yday

    @property
    def hours(self) -> int:
        """ hours, range [0, 23] """
        return self.localtime.tm_hour

    @property
    def minutes(self) -> int:
        """ minutes, range [0, 59] """
        return self.localtime.tm_min

    @property
    def seconds(self) -> int:
        """ seconds, range [0, 61]) """
        return self.localtime.tm_sec

    #
    #  Factory
    #

    @classmethod
    def parse(cls, value: Any):
        """ parse timestamp value (seconds from 1970-01-01 00:00:00) """
        if value is None:
            return None
        elif isinstance(value, DateTime):
            # exactly
            return value
        elif isinstance(value, float):
            # assuming it is a timestamp value in seconds
            return DateTime(timestamp=value)
        elif isinstance(value, int):
            # assuming it is a timestamp value in seconds
            return DateTime(timestamp=float(value))
        else:
            raise TypeError('invalid time value: %s' % value)

    @classmethod
    def now(cls):
        seconds = time.time()
        return DateTime(timestamp=seconds)

    @classmethod
    def current_timestamp(cls) -> float:
        """ Return the current time in seconds since the Epoch. """
        return time.time()

    @classmethod
    def sleep(cls, seconds: Union[float, int]):
        time.sleep(seconds)



# -*- coding: utf-8 -*-
# ==============================================================================
# MIT License
#
# Copyright (c) 2020 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from abc import ABC, abstractmethod
from typing import Optional


class StringCoder(ABC):
    """
        String Coder
        ~~~~~~~~~~~~
        UTF-8, UTF-16, GBK, GB2312, ...

        1. encode string to binary data;
        2. decode binary data to string.
    """

    @abstractmethod
    def encode(self, string: str) -> bytes:
        """
        Encode local string to binary data

        :param string: local string
        :return: binary data
        """
        raise NotImplemented

    @abstractmethod
    def decode(self, data: bytes) -> Optional[str]:
        """
        Decode binary data to local string

        :param data: binary data
        :return: local string
        """
        raise NotImplemented


class UTF8:
    coder: StringCoder = None

    @staticmethod
    def encode(string: str) -> bytes:
        # assert UTF8.coder is not None, 'UTF8 parser not set yet'
        return UTF8.coder.encode(string=string)

    @staticmethod
    def decode(data: bytes) -> Optional[str]:
        # assert UTF8.coder is not None, 'UTF8 parser not set yet'
        return UTF8.coder.decode(data=data)


#
#   Interfaces
#


def utf8_encode(string: str) -> bytes:
    return UTF8.encode(string=string)


def utf8_decode(data: bytes) -> Optional[str]:
    return UTF8.decode(data=data)
